package vsl.dataflow;

import java.io.*;
import javax.imageio.*;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;

import org.w3c.dom.*;
import vsl.dataflow.util.*;

/*
 * Abstrakte Basisklasse fr alle Knoten eines Datenflussdiagramms.
 */
public abstract class DataFlowNode extends DataFlowComponent
{

  /*
   * Mgliche Ausprgungen
   */
  public static final int DF_OTHER_NODE = -1;
  public static final int DF_DIAGRAM_NODE = 1;
  public static final int DF_CONSTRUCTOR_NODE = 2;
  public static final int DF_CONSTANT_NODE = 3;
  public static final int DF_FUNCTION_NODE = 4;
  public static final int DF_STATE_NODE = 5;
  public static final int DF_OPERATOR_NODE = 6;

  /*
   * Knoten sind durch ihren Namen und ihre ID (beides Strings) innerhalb eines
   * Datenflussdiagramms eindeutig identifizierbar.
   */
  protected String name;
  protected String id = new String( "00" );

  /*
   * Liste der Eingnge des Knotens
   */
  protected DataFlowAnchorList inputs = null;

  /*
   * Liste der Ausgnge des Knotens
   */
  protected DataFlowAnchorList outputs = null;

  /*
   * Maximale Entfernung des Knotens zur Senke
   */
  protected int maxDistanceToRoot = -1;

  /*
   * Hilfsvariable fr Berechnungen innerhalb des Diagramms
   */
  private boolean marked = false;

  /*
   * Jeder Knoten muss sich als DOM Element exportieren knnen.
   */
  public abstract Element exportXML( Document document );

  /*
   * Gibt den Knotentyp zurueck.
   */
  public abstract int getNodeType();

  /*
   * Beim Verschieben mssen die Positionen der Ein- und Ausgnge
   * aktualisert werden. Deswegen wurde setLocation berschrieben.
   */

 public void setLocation( Point p )
  {
    super.setLocation(p);
    updateAnchorLocations();
  }

  /*
   * Beim Verschieben mssen die Positionen der Ein- und Ausgnge
   * aktualisert werden. Deswegen wurde setLocation berschrieben.
   */
  public void setLocation( int x, int y )
  {
    super.setLocation( x, y );
    updateAnchorLocations();
  }

  /*
   * Richtet die Ein- und Ausgnge aus.
   */
  protected void updateAnchorLocations()
  {
    // Eingnge
    double inputDistance = ( double ) getHeight() / ( ( double ) ( inputs.getAnchorCount()+1 ) );
    for ( int i = 0; i < inputs.getAnchorCount(); i++ )
    {
      DataFlowAnchor input = inputs.getAnchor( i );

      input.setLocation( new Point( getX() - input.getWidth(),
                                    getY() - input.getHeight()/2 +
                                    ( int ) ( inputDistance * ( double ) ( i+ 1 ) ) ) );
    }
    // Ausgnge
    double outputDistance = ( double ) getHeight() / ( double ) ( outputs.getAnchorCount() + 1 );
    for ( int i = 0; i < outputs.getAnchorCount(); i++ )
    {
      DataFlowAnchor output = outputs.getAnchor( i );
      output.setLocation( getX() + getWidth(),
                          getY() - output.getHeight() / 2 +
                          ( int ) ( outputDistance * ( double ) ( i + 1 ) ) );
    }
  }

  /*
   * Die Namen der Ein- und Ausgnge anzeigen!
   */
  public void showLabels()
  {
    for ( int i = 0; i < inputs.getAnchorCount(); i++ )
    {
      inputs.getAnchor( i ).setShowLabel( true );
    }
    for ( int i = 0; i < outputs.getAnchorCount(); i++ )
    {
      outputs.getAnchor( i ).setShowLabel( true );
    }
  }

  /*
   * Die Namen der Ein- und Ausgnge ausblenden!
   */
  public void hideLabels()
  {
    for ( int i = 0; i < inputs.getAnchorCount(); i++ )
    {
      inputs.getAnchor( i ).setShowLabel( false );
    }
    for ( int i = 0; i < outputs.getAnchorCount(); i++ )
    {
      outputs.getAnchor( i ).setShowLabel( false );
    }
  }

  /*
   * Berechnet rekursiv die Entfernung des Knotens zur Datensenken des
   * Diagramms.
   */
  protected int calculateMaxDistance( int depth )
  {
    int max = depth;
    if ( depth > maxDistanceToRoot )
    {
      maxDistanceToRoot = depth;
    }
    for ( int i = 0; i < inputs.getAnchorCount(); i++ )
    {
      DataFlowDestinationAnchor input = ( DataFlowDestinationAnchor ) inputs.getAnchor( i );
      if ( input.isConnected() )
      {
        DataFlowNode node = input.getEdge().getSource().getNode();
        if ( node != null )
        {
          int tmp = node.calculateMaxDistance( depth + 1 );
          if ( tmp > max )
          {
            max = tmp;
          }
        }
      }
    }
    return max;
  }

  /*
   * Gibt wahr zurck, wenn innerhalb eines Diagrammes eine Zyklus ist.
   */
  protected boolean detectCycle()
  {
    if ( marked )
    {
      marked = false;
      return true;
    }
    else
    {
      marked = true;
      boolean detected = false;
      for ( int i = 0; i < inputs.getAnchorCount(); i++ )
      {
        DataFlowDestinationAnchor input = ( DataFlowDestinationAnchor ) inputs.getAnchor( i );
        if ( input.isConnected() )
        {
          DataFlowNode node = input.getEdge().getSource().getNode();
          if ( node != null )
          {
            if ( node.detectCycle() )
            {
              marked = false;
              return true;
            }
          }
        }
      }
      marked = false;
      return false;
    }
  }

  /*
   * Gibt die maximale Entfernung des Knotens zur Datensenke zurck.
   */
  public int getMaxDistance()
  {
    return maxDistanceToRoot;
  }

  /*
   * Gibt Knoteninformationen in die Konsole aus.
   */
  public void print()
  {
    System.out.println( "" );
    System.out.println( "Node " + name + " with id " + id );
    System.out.println( "Maximum distance to root is " + maxDistanceToRoot );
    for ( int i = 0; i < inputs.getAnchorCount(); i++ )
    {
      DataFlowDestinationAnchor input = ( DataFlowDestinationAnchor ) inputs.getAnchor( i );
      if ( input.isConnected() )
      {
        System.out.println( "Anchor " + inputs.getAnchor( i ).getName()
                            + " is conntected to Anchor " +
                            input.getEdge().getSource().getName()
                            + " with Anchor " + input.getEdge().getSource().getName() );
      }
    }
    for ( int i = 0; i < inputs.getAnchorCount(); i++ )
    {
      DataFlowDestinationAnchor input = ( DataFlowDestinationAnchor ) inputs.getAnchor( i );
      if ( input.isConnected() )
      {
        if ( input.getEdge().getSource().hasNode() )
        {
          input.getEdge().getSource().getNode().print();
        }
      }
    }
  }

  /*
   * Gibt die Anzahl der Eingnge zurck.
   */
  public int getInputCount()
  {
    return inputs.getAnchorCount();
  }

  /*
   * Gibt die Anzahl der Ausgnge zurck.
   */
  public int getOutputCount()
  {
    return outputs.getAnchorCount();
  }

  public DataFlowDestinationAnchor getInputAnchor( String anchorName )
  {
    return ( DataFlowDestinationAnchor ) inputs.getAnchor( anchorName );
  }

  public DataFlowDestinationAnchor getInputAnchor( int i )
  {
    return ( DataFlowDestinationAnchor ) inputs.getAnchor( i );
  }

  public DataFlowSourceAnchor getOutputAnchor( String anchorName )
  {
    return ( DataFlowSourceAnchor ) outputs.getAnchor( anchorName );
  }

  public DataFlowSourceAnchor getOutputAnchor( int i )
  {
    return ( DataFlowSourceAnchor ) outputs.getAnchor( i );
  }

  /*
   * Setzt den Namen des Knotens.
   */
  public void setName( String name )
  {
    this.name = name;
  }

  /*
   * Gibt den Namen des Knotens zurck.
   */
  public String getName()
  {
    return name;
  }

  /*
   * Setzt die ID des Knotens.
   */
  public void setId( String id )
  {
    this.id = id;
  }

  /*
   * Gibt die ID des Knotens zurck.
   */
  public String getId()
  {
    return id;
  }

  public int getComponentType()
  {
    return DF_NODE;
  }

  public boolean isSelectable()
  {
    return true;
  }

  public boolean isDraggable()
  {
    return true;
  }

}